/**
 * native libraries for custom architectures
 */
module com.kohlschutter.junixsocket.nativecustom {
  exports org.newsclub.lib.junixsocket.custom;

  requires static com.kohlschutter.annotations.compiletime;
}
