/**
 * Helper package to identify the Maven artifact with JNI libraries for common architectures.
 *
 * The set of common architectures is configured through Maven profiles and should at least contain
 * versions for Intel 64-bit macOS and Linux.
 */
package org.newsclub.lib.junixsocket.common;
