/**
 * native libraries for common architectures
 */
module com.kohlschutter.junixsocket.nativecommon {
  exports org.newsclub.lib.junixsocket.common;
  requires static com.kohlschutter.annotations.compiletime;
}
