/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_newsclub_net_unix_android_SocketOutputStream */

#ifndef _Included_org_newsclub_net_unix_android_SocketOutputStream
#define _Included_org_newsclub_net_unix_android_SocketOutputStream
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_newsclub_net_unix_android_SocketOutputStream
 * Method:    socketWrite0
 * Signature: (Ljava/io/FileDescriptor;[BII)V
 */
JNIEXPORT void JNICALL Java_org_newsclub_net_unix_android_SocketOutputStream_socketWrite0
  (JNIEnv *, jobject, jobject, jbyteArray, jint, jint);

#ifdef __cplusplus
}
#endif
#endif
