/*
 * junixsocket
 *
 * Copyright 2009-2024 Christian Kohlschütter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.newsclub.net.unix.demo.rmi;

import java.rmi.RemoteException;

import org.newsclub.net.unix.demo.rmi.services.World;

/**
 * The implementation of the very simple {@link World} service.
 *
 * @author Christian Kohlschütter
 */
public class WorldImpl implements World {
  private final String text;

  /**
   * Creates a new {@link World} instance.
   *
   * @param text The text to return upon calling {@link #world()}.
   */
  public WorldImpl(String text) {
    this.text = text;
  }

  @Override
  public String world() throws RemoteException {
    return text;
  }
}
