/*
 * junixsocket
 *
 * Copyright 2009-2024 Christian Kohlschütter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.newsclub.net.unix.pool;

import org.eclipse.jdt.annotation.Nullable;

/**
 * Holds an object.
 *
 * @param <O> The object type.
 * @author Christian Kohlschütter
 */
public final class MutableHolder<O> {
  private @Nullable O obj;

  /**
   * Creates a new {@link MutableHolder} object, holding the given object.
   *
   * @param obj The object to hold, or {@code null}.
   */
  public MutableHolder(@Nullable O obj) {
    this.obj = obj;
  }

  /**
   * Returns the current object.
   *
   * @return The object.
   */
  public @Nullable O get() {
    return obj;
  }

  /**
   * Sets the object.
   *
   * @param obj The object.
   */
  public void set(@Nullable O obj) {
    this.obj = obj;
  }
}
