/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_newsclub_net_unix_android_SocketInputStream */

#ifndef _Included_org_newsclub_net_unix_android_SocketInputStream
#define _Included_org_newsclub_net_unix_android_SocketInputStream
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_newsclub_net_unix_android_SocketInputStream
 * Method:    socketRead0
 * Signature: (Ljava/io/FileDescriptor;[BIII)I
 */
JNIEXPORT jint JNICALL Java_org_newsclub_net_unix_android_SocketInputStream_socketRead0
  (JNIEnv *, jobject, jobject, jbyteArray, jint, jint, jint);

#ifdef __cplusplus
}
#endif
#endif
