/*
 * junixsocket
 *
 * Copyright 2009-2024 Christian Kohlschütter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.newsclub.net.unix;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Declares some operation "unsafe", which is guarded via
 * {@link AFSocketCapability#CAPABILITY_UNSAFE}.
 * <p>
 * Methods annotated with this type should also call {@link AFSocket#ensureUnsafeSupported()} to
 * check availability at runtime.
 * <p>
 * <b>NOTE:</b> This annotation has nothing to do with {@code sun.misc.Unsafe}.
 *
 * @author Christian Kohlschütter
 * @see AFSocket#ensureUnsafeSupported()
 */
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.METHOD)
public @interface Unsafe {

}
